---- DATA STRUCTURES ----
--https://developer.roblox.com/en-us/articles/Table#:~:text=An%20item%20can%20be%20removed,items%20down%20one%20index%20position.&text=For%20more%20details%20on%20the,the%20Lua%20table%20library%20documentation.

--[[
    --REFER to Monkey_Cmd.lua to send command
    --SEND THIS DATA WITH SCOPE "Monkey_EventScheduler" and COMMAND "SetScheduledEvent" with the following args data structure:

local eventData = {
    Interval = "", --OneMinute,TenMinutes,Hours,Days --Event will be updated every specified interval.

    Scope = "", --String --This is the scope that will be used to send the command OnEventStart, OnEventFinish.
    Name = "", --String --Name of event

    DeleteOnStart = false,

    StartEvent =  --In how many minutes/tens of minutes/hours/days?
        {
            Value = 0, --Int
            Deviation = 0, --Int How much will the Value randomly deviate by (+/-)
            OnValue =
            {
                [1] =
                {
                    Command = OnEventCountdownCommand,
                    Args =
                    {
                        string = 'One Hour remaining!'
                    }
                }
            }
        },

    EventLength =  --OPTIONAL --How long will the event be in minutes/tens of minutes/hours/days? 
        {
            Value = 0, --Int
            Deviation = 0 --Int How much will the Value randomly deviate by (+/-)
        },

    OnEventStart = 
        {
            Command = "", --String --Which server command to send when StartEvent time is reached?
            Args = nil, --Optional data --Optional args to send with command.
        },
    OnEventFinish = --OPTIONAL
        {
            Command = "", --String --Which server command to send when StartEvent time is reached?
            Args = nil, --Optional data --Optional args to send with command.
        },
}

--]]

---- INITIALISATION CHECK ----

if not (Monkey_Utils:IsSingleplayer() or isServer() or isCoopHost()) then
    return;
end

---- IMPORTS ----
local Constants = require "EventScheduler/Monkey_EventSchedulerConstants";
local LibConstants = require "MonkeyLibrary_Constants";

local Monkey_ModData = require 'ModDataManager/Monkey_ModData';
local ModData;

local Monkey_Logger = require 'Logging/Monkey_Logger';
local Log = Monkey_Logger:new('MK_EventScheduler');

Log:LogInfo('Initialised Monkey_EventScheduler!');

---- PRIVATE FUNCTIONS ----

local function IsEventValid(event)
    if not event then
        Log:LogError('VALIDATION: Event is NIL!');
        return false;
    end

    if not event.Interval or not event.Scope or not event.Name then
        Log:LogError('VALIDATION: Interval, Scope or Name are NIL!');
        return false;
    end

    if not event.StartEvent then
        Log:LogError('VALIDATION: StartEvent is NIL!');
        return false;
    end

    return true;
end

local function getTimeEventValue(eventLength)
    if not eventLength then
        return nil;
    end

    local startEventInVar = math.abs(eventLength.Value);

    if eventLength.Deviation then
        local deviation = math.abs(eventLength.Deviation);
        if deviation > 0 then
            local rand = ZombRand(deviation) + 1;
            if ZombRand(2) == 0 then
                rand = rand * -1;
            end

            startEventInVar = startEventInVar + rand;
        end
    end

    return startEventInVar;
end

local function initialise()
    if not ModData:GetMonkeyModData().loaded then
        ModData:GetMonkeyModData().loaded = true;
        local modData = ModData:GetMonkeyModData();
        modData.intervals = 
        {
            OneMinute = 
            {
                Count = 0,
                Events = {}
            },
            TenMinutes =
            {
                Count = 0,
                Events = {}
            },
            Hours =
            {
                Count = 0,
                Events = {}
            },
            Days = 
            {
                Count = 0,
                Events = {}
            }
        }

        Log:LogInfo('Initialised Monkey_EventScheduler data structure!');
    end
end

--[[
    Checks for duplicate events on Scope and Name
--]]
local function eventExists(tbl, scope, name)
    --check event table for duplicate scope and name
    --this should be done before adding an event
    if tbl.Count == 0 then
        return false;
    end
        
    for i = 1, tbl.Count do
        local event = tbl.Events[i];
        if event.Scope == scope and event.Name == name then
            return true;
        end
    end

    return false;
end

--[[
    Deletes event by matching Scope and Name
--]]
local function deleteEvent(tbl, event)
    for i = 1, tbl.Count do
        local tempEvent = tbl.Events[i];
        if tempEvent.Scope == event.Scope and tempEvent.Name == event.Name then
            Log:LogInfo('Deleted event with Scope: ' .. tempEvent.Scope .. ' & Name: ' .. tempEvent.Name .. ' at Index: ' .. i .. '!' );
            table.remove(tbl.Events, i);
            if tbl.Count ~= nil then
                tbl.Count = tbl.Count - 1;
            end
            return tbl;
        end
    end
end

--[[ 
    Returns the event if it should be deleted
--]]
local function handleEvent(event)
    if event.StartEventIn > 0 then
        --Tick down event counter
        event.StartEventIn = event.StartEventIn - 1;

        --Handle event countdown
        if event.EventStartCountdown and #event.EventStartCountdown > 0 then
            local eventCountdown = event.EventStartCountdown[event.StartEventIn];
            if eventCountdown then
                Monkey_Cmd:SendCommand(event.Scope, eventCountdown.Command, eventCountdown.Args);
            end
        end

        --If counter is ticked down, trigger command if available
        if event.StartEventIn <= 0 then
            if event.OnEventStart then
                Monkey_Cmd:SendCommand(event.Scope, event.OnEventStart.Command, event.OnEventStart.Args);
            end
            if event.DeleteOnStart then
                return event; 
            end
        end
    elseif event.EndEventIn ~= nil then
        --Tick down event counter
        if event.EndEventIn > 0 then
            event.EndEventIn = event.EndEventIn - 1;

            --Handle event countdown
            if event.EventFinishCountdown and #event.EventFinishCountdown > 0 then
                local eventCountdown = event.EventFinishCountdown[event.EndEventIn];
                if eventCountdown then
                    Monkey_Cmd:SendCommand(event.Scope, eventCountdown.Command, eventCountdown.Args);
                end
            end

            --If counter is ticked down, trigger command if available
            if event.EndEventIn <= 0 then
                if event.OnEventFinish then
                    Monkey_Cmd:SendCommand(event.Scope, event.OnEventFinish.Command, event.OnEventFinish.Args);
                end
                return event;
            end
        else
            return event;
        end
    else
        return event;
    end

    return nil;
end

--[[ 
    Iterates through each event and handles it
--]]
local function handleTickEvents(interval)
    local modData = ModData:GetMonkeyModData();
    local events = modData.intervals[interval];
      
    local eventsToDelete = {
        Count = 0,
        Events = {},
    };

    if not events then
        return;
    end

    local eventResult;
    local tempEvent;

    for i = 1, events.Count do
        tempEvent = events.Events[i];
        eventResult = handleEvent(tempEvent);
        if eventResult then
            eventsToDelete.Count = eventsToDelete.Count + 1;
            table.insert(eventsToDelete.Events, eventResult);
        end
    end

    if eventsToDelete.Count > 0 then
        for i = 1, eventsToDelete.Count do
            local tempEvent = eventsToDelete.Events[i];
            events = deleteEvent(events, tempEvent);
        end
    end

    modData.intervals[interval] = events;
end

---- PUBLIC FUNCTIONS ----

Monkey_EventScheduler = {}

function Monkey_EventScheduler:SetScheduledEvent(eventData)
    if not IsEventValid(eventData) then
        Log:LogError('Scheduled event is invalid!');
        return false;
    end

    local modData = ModData:GetMonkeyModData();
    local events = modData.intervals[eventData.Interval];

    if eventExists(events, eventData.Scope, eventData.Name) then
        Log:LogError('An event with Scope: ' .. eventData.Scope .. ' & Name: ' .. eventData.Name .. ' already exists!');
        return false;
    end

    events.Count = events.Count + 1;
    events.Events[events.Count] = 
    {
        ScheduledByPlayerOnlineID = eventData.PlayerOnlineID,
        Scope = eventData.Scope,
        Name = eventData.Name,
        StartEventIn = Monkey_Utils:GetValueWithDeviation(eventData.StartEvent),
        EndEventIn = Monkey_Utils:GetValueWithDeviation(eventData.EventLength),
        OnEventStart = eventData.OnEventStart,
        EventStartCountdown = eventData.StartEvent.OnValue,
        OnEventFinish = eventData.OnEventFinish,
        EventFinishCountdown = eventData.EventLength.OnValue,
        DeleteOnStart = eventData.DeleteOnStart
    };

    Log:LogInfo('Added new Event with Scope: ' .. eventData.Scope .. ' & Name: ' .. eventData.Name .. '!');

    return true;
end

---- CHECK PER INTERVAL ----

local function checkOneMinuteEvents()
    handleTickEvents(Constants.OneMinute);
end

local function checkTenMinutesEvents()
    handleTickEvents(Constants.TenMinutes);
end

local function checkHoursEvents()
    handleTickEvents(Constants.Hours);
end

local function checkDaysEvents()
    handleTickEvents(Constants.Days);
end

Events.EveryOneMinute.Add(checkOneMinuteEvents);
Events.EveryTenMinutes.Add(checkTenMinutesEvents);
Events.EveryHours.Add(checkHoursEvents);
Events.EveryDays.Add(checkDaysEvents);

---- SERVER COMMAND INTEGRATION ----

local function onServerCommand(scopeName, command, args)
	if scopeName ~= Constants.EventSchedulerScope or command ~= Constants.SetScheduledEventCommand then 
        return 
    end;

    if args == nil then
        return;
    end

    Log:LogInfo('Recieved ' .. scopeName .. '-' .. command .. '!');

    Monkey_EventScheduler:SetScheduledEvent(args);
end

---- INITIALISATION ----

local function onGameStart_initialise()
    Events.OnServerCommand.Add(onServerCommand);
    ModData = Monkey_ModData:init(LibConstants.ModName, Constants.EventSchedulerScope);
    initialise();
end

if isServer() then
    Events.OnServerStarted.Add(onGameStart_initialise);
else
    Events.OnGameStart.Add(onGameStart_initialise);
end